"use strict";
/**
 * @license
 * Copyright (c) 2016 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
exports.__esModule = true;
var browser_capabilities_1 = require("browser-capabilities");
var content_type_1 = require("content-type");
var fs = require("fs");
var LRU = require("lru-cache");
var path = require("path");
var polymer_build_1 = require("polymer-build");
var get_compile_target_1 = require("./get-compile-target");
var transform_middleware_1 = require("./transform-middleware");
var javaScriptMimeTypes = [
    'application/javascript',
    'application/ecmascript',
    'text/javascript',
    'text/ecmascript',
];
var htmlMimeType = 'text/html';
var compileMimeTypes = [htmlMimeType].concat(javaScriptMimeTypes);
// Match the polyfills from https://github.com/webcomponents/webcomponentsjs,
// but not their tests.
exports.isPolyfill = /(^|\/)webcomponentsjs\/[^\/]+$/;
function getContentType(response) {
    var contentTypeHeader = response.get('Content-Type');
    return contentTypeHeader && content_type_1.parse(contentTypeHeader).type;
}
// NOTE: To change the max length of the cache at runtime, just use bracket
// notation, i.e. `babelCompileCache['max'] = 64 * 1024` for 64KB limit.
exports.babelCompileCache = LRU({
    length: function (n, key) { return n.length + key.length; }
});
// TODO(justinfagnani): see if we can just use the request path as the key
// See https://github.com/Polymer/polyserve/issues/248
exports.getCompileCacheKey = function (requestPath, body, options) {
    return JSON.stringify(options) + requestPath + body;
};
function babelCompile(compile, moduleResolution, rootDir, packageName, componentUrl, componentDir) {
    return transform_middleware_1.transformResponse({
        shouldTransform: function (request, response) {
            // We must never compile the Custom Elements ES5 Adapter or other
            // polyfills/shims.
            if (exports.isPolyfill.test(request.url)) {
                return false;
            }
            if ('nocompile' in request.query) {
                return false;
            }
            if (!compileMimeTypes.includes(getContentType(response))) {
                return false;
            }
            if (compile === 'always' || moduleResolution === 'node') {
                return true;
            }
            if (compile === 'never') {
                return false;
            }
            var capabilities = browser_capabilities_1.browserCapabilities(request.get('user-agent'));
            return !capabilities.has('es2015') || !capabilities.has('modules');
        },
        transform: function (request, response, body) {
            var capabilities = browser_capabilities_1.browserCapabilities(request.get('user-agent'));
            var compileTarget = get_compile_target_1.getCompileTarget(capabilities, compile);
            var options = {
                compileTarget: compileTarget,
                transformModules: compile === 'always' || !capabilities.has('modules'),
            };
            var cacheKey = exports.getCompileCacheKey(request.baseUrl + request.path, body, options);
            var cached = exports.babelCompileCache.get(cacheKey);
            if (cached !== undefined) {
                return cached;
            }
            // Make sure that componentDir is absolute, like jsTransform expects
            componentDir = path.resolve(rootDir, componentDir);
            var transformed;
            var contentType = getContentType(response);
            var requestPath = request.path;
            // TODO(justinfagnani): Use path-is-inside, but test on Windows
            var isRootPathRequest = request.path === "/" + packageName ||
                request.path.startsWith("/" + packageName + "/");
            var isComponentRequest = request.baseUrl === "/" + componentUrl;
            var filePath;
            if (isRootPathRequest) {
                requestPath = requestPath.substring(("/" + packageName).length);
            }
            if (isComponentRequest && !isRootPathRequest) {
                filePath = path.join(componentDir, requestPath);
            }
            else {
                filePath = path.join(rootDir, requestPath);
            }
            // The file path needs to include the filename for correct relative
            // path calculation
            try {
                var stat = fs.statSync(filePath);
                if (stat.isDirectory()) {
                    filePath = path.join(filePath, 'index.html');
                }
            }
            catch (e) {
                // file not found, will 404 later
            }
            if (contentType === htmlMimeType) {
                transformed = polymer_build_1.htmlTransform(body, {
                    js: {
                        compile: options.compileTarget,
                        moduleResolution: moduleResolution,
                        filePath: filePath,
                        packageName: packageName,
                        isComponentRequest: isComponentRequest,
                        componentDir: componentDir,
                        rootDir: rootDir,
                        transformModulesToAmd: options.transformModules,
                        softSyntaxError: true,
                    },
                    injectAmdLoader: options.transformModules,
                });
            }
            else if (javaScriptMimeTypes.includes(contentType)) {
                transformed = polymer_build_1.jsTransform(body, {
                    compile: options.compileTarget,
                    transformModulesToAmd: options.transformModules ? 'auto' : false,
                    moduleResolution: moduleResolution,
                    filePath: filePath,
                    isComponentRequest: isComponentRequest,
                    packageName: packageName,
                    componentDir: componentDir,
                    rootDir: rootDir,
                });
            }
            else {
                transformed = body;
            }
            exports.babelCompileCache.set(cacheKey, transformed);
            return transformed;
        },
    });
}
exports.babelCompile = babelCompile;
