"use strict";
/**
 * @license
 * Copyright (c) 2015 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
exports.__esModule = true;
var assert = require("assert");
var fs = require("fs");
var mime = require("mime");
var path = require("path");
/** h2 push manifest cache */
var _pushManifest = {};
/**
 * Asserts file existence for all specified files in a push-manifest
 * @param root path to root directory
 * @param manifest manifest object
 */
function assertValidManifest(root, manifest) {
    function assertExists(filename) {
        var fname = path.join(root, filename);
        try {
            // Ignore root path, since that always exists for the router
            if (filename !== '/') {
                assert(fs.existsSync(fname), "not found: " + fname);
            }
        }
        catch (err) {
            throw new Error("invalid h2-push manifest: " + err);
        }
    }
    for (var _i = 0, _a = Object.keys(manifest); _i < _a.length; _i++) {
        var refFile = _a[_i];
        assertExists(refFile);
        for (var _b = 0, _c = Object.keys(manifest[refFile]); _b < _c.length; _b++) {
            var pushFile = _c[_b];
            assertExists(pushFile);
        }
    }
}
/**
 * Reads a push-manifest from the specified path, or a cached version
 * of the file
 * @param root path to root directory
 * @param manifestPath path to manifest file
 * @returns the manifest
 */
function getPushManifest(root, manifestPath) {
    if (!_pushManifest[manifestPath]) {
        var data = fs.readFileSync(manifestPath);
        var manifest = JSON.parse(data.toString());
        assertValidManifest(root, manifest);
        _pushManifest[manifestPath] = manifest;
    }
    return _pushManifest[manifestPath];
}
exports.getPushManifest = getPushManifest;
/**
 * Pushes any resources for the requested file
 * @param options server options
 * @param req HTTP request
 * @param res HTTP response
 */
function pushResources(options, req, res) {
    if (res.push && options.protocol === 'h2' && options.pushManifestPath &&
        !req.get('x-is-push')) {
        // TODO: Handle preload link headers
        var pushManifest = getPushManifest(options.root, options.pushManifestPath);
        var resources = pushManifest[req.path];
        if (resources) {
            var root = options.root;
            var _loop_1 = function (filename) {
                var stream = res.push(filename, {
                    request: { accept: '*/*' },
                    response: {
                        'content-type': mime.getType(filename),
                        // Add an X-header to the pushed request so we
                        // don't trigger pushes for pushes
                        'x-is-push': 'true'
                    }
                })
                    .on('error', function (err) { return console.error('failed to push', filename, err); });
                fs.createReadStream(path.join(root, filename)).pipe(stream);
            };
            for (var _i = 0, _a = Object.keys(resources); _i < _a.length; _i++) {
                var filename = _a[_i];
                _loop_1(filename);
            }
        }
    }
}
exports.pushResources = pushResources;
